from typing import List, Optional, Union
import glob
from pydantic import ValidationError
import requests
from requests import Response
import json
import os, sys
import eyed3
from mutagen.mp4 import MP4, MP4Cover
import logging

from . import common
from .models import itunes_api, modes

logger = logging.getLogger(__name__)


def itunes_lib_search(
    itunes_lib_path: str,
    search_parameters: str,
    album_properties: Optional[itunes_api.ItunesApiAlbumKeys] = None,
) -> List[str]:
    """
    Performs a search on users iTunes library by album, artist and genre

    Args:
        itunes_lib_path (str): path to itunes song library on disk
        search_parameters (str): search term
        album_properties (itunes_api.ItunesApiAlbumKeys): determines whether to do a smarter search based on given album properties

    Returns: a list of songpaths found
    """
    itunes_songs = glob.glob(
        os.path.join(itunes_lib_path, "*", "*", f"*.*"), recursive=True
    )
    matches = []
    for song_path in itunes_songs:
        # song_name_split is list of itunes file path.. artist is -3 from length, song is -1
        song_name_split = song_path.split(os.sep)
        song_name = common.remove_illegal_characters(
            song_name_split[len(song_name_split) - 1].lower()
        )
        album_name = common.remove_illegal_characters(
            song_name_split[len(song_name_split) - 2].lower()
        )
        artist_name = common.remove_illegal_characters(
            song_name_split[len(song_name_split) - 3].lower()
        )
        search_parameters = common.remove_illegal_characters(search_parameters.lower())
        if album_properties == None:
            formatted_name = song_name + " " + album_name + " " + artist_name
            if search_parameters in formatted_name:
                matches.append(song_path)
        else:
            if (
                album_properties.artistName.lower() in artist_name
                and search_parameters in song_name
            ):
                matches.append(song_path)

    # returns sorted alphabetical list of matches.
    matches.sort()
    return matches


def m4a_tagger(file_path: str, song_tag_data: itunes_api.ItunesApiSongModel) -> bool:
    """Tag an m4a file using iTunes recognized tags.

    The below is a tag legend taken from mutagen website.
    Text values:

    ```
    '\\xa9nam' - track title
    '\\xa9alb' - album
    '\\xa9ART' - artist
    'aART' - album artist
    '\\xa9day' - year
    'purd' - purchase date
    '\\xa9gen' - genre
    Tuples of ints (multiple values per key are supported):
    'trkn' - track number, total tracks
    'disk' - disc number, total discs
    'covr' - cover artwork, list of MP4Cover objects (which are tagged strs)
    ```

    Args:
        file_path (str): path to the m4a file
        song_tag_data (itunes_api.ItunesApiSongModel): The model for the songs relevant metadata collected from itunes search api

    Returns:
        bool: true if tagging was successful.
    """
    try:
        logger.info(f"Adding tags to m4a file : {file_path}")

        response = artwork_searcher(url=song_tag_data.artworkUrl100)
        audiofile = MP4(file_path)

        # Set all the tags for the mp3, all without if statement were checked for existence.
        audiofile["\xa9ART"] = song_tag_data.artistName
        audiofile["\xa9alb"] = song_tag_data.collectionName
        audiofile["\xa9nam"] = song_tag_data.trackName
        audiofile["\xa9gen"] = song_tag_data.primaryGenreName
        audiofile["trkn"] = [(song_tag_data.trackNumber, song_tag_data.trackCount)]
        audiofile["disk"] = [(song_tag_data.discNumber, song_tag_data.discCount)]
        audiofile["\xa9day"] = song_tag_data.releaseDate

        if (
            song_tag_data.collectionArtistName is not None
        ):  # check if collection_artist_name exists before adding to tags
            audiofile["aART"] = song_tag_data.collectionArtistName

        if response.status_code == 200:
            audiofile["covr"] = [
                MP4Cover(response.content, imageformat=MP4Cover.FORMAT_PNG)
            ]

        audiofile.save()
        logger.info("Your tags have been set.")
        return True
    except Exception as e:
        logger.exception(
            f"Unexpected error occured while trying to tag your m4a file: {e}"
        )
        return False


def mp3ID3Tagger(mp3_path: str, song_tag_data: itunes_api.ItunesApiSongModel) -> bool:
    """
    Tags an mp3 file at mp3_path given a itunes_api.ItunesApiSongModel object

    Args:
        mp3_path (str): file path of mp3 file
        song_tag_data (itunes_api.ItunesApiSongModel): the model with tag fields

    Returns:
        bool: true if tagging was a success
    """
    try:
        # Create MP3File instance.
        logger.info(f"Adding your tags to mp3 file: {mp3_path}")
        # Have to call MP3File twice for it to work.

        # Get the image to show for a song .. but get high res
        # get album artwork from the list of sizes

        response = artwork_searcher(url=song_tag_data.artworkUrl100)

        # Set all the tags for the mp3, all without if statement were checked for existence.
        audiofile = eyed3.load(mp3_path)
        audiofile.tag.artist = song_tag_data.artistName
        audiofile.tag.album = song_tag_data.collectionName
        audiofile.tag.title = song_tag_data.trackName
        audiofile.tag.genre = song_tag_data.primaryGenreName
        audiofile.tag.track_num = (song_tag_data.trackNumber, song_tag_data.trackCount)
        audiofile.tag.disc_num = (song_tag_data.discNumber, song_tag_data.discCount)
        audiofile.tag.recording_date = song_tag_data.releaseDate

        if (
            song_tag_data.collectionArtistName is not None
        ):  # check if collection_artist_name exists before adding to tags
            audiofile.tag.album_artist = song_tag_data.collectionArtistName

        if response.status_code == 200:
            audiofile.tag.images.set(
                type_=3,
                img_data=response.content,
                mime_type="image/png",
                description="Art",
                img_url=None,
            )

        audiofile.tag.save()
        logger.info("Your tags have been set.")
        return True

    except Exception as e:
        logger.exception(
            f"Unexpected error occured while trying to tag your mp3 file: {e}"
        )
        return False


def query_api(
    search_variable: str, limit: int, mode: modes.Modes, lookup: bool = False
) -> List[Union[itunes_api.ItunesApiSongModel, itunes_api.ItunesApiAlbumKeys]]:
    """
    Args:
        search_variable (str): the term to search itunes api for
        limit  (int): limit of the search in the api
        mode (modes.Modes): value of album, or song for now
        lookup (bool): if true, perform a lookup search via itunes api rather than a default search
    Returns
        List[Union[itunes_api.ItunesApiSongModel, itunes_api.ItunesApiAlbumKeys]]: a list containing a
            tuple with itunes song properties and album properties
    """
    parsed_results_list = []
    result_dict = {}
    if not lookup:  # perform general search
        search_parameters = {
            "term": search_variable,
            "entity": mode.value,
            "limit": limit,
        }
        itunes_response = requests.get(
            "https://itunes.apple.com/search", params=search_parameters
        )
    else:  # perform lookup query by itunes id
        search_parameters = {
            "id": search_variable,
            "entity": mode.value,
            "limit": limit,
        }
        itunes_response = requests.get(
            "https://itunes.apple.com/lookup", params=search_parameters
        )

    # itunesResponse = requests.get('https://itunes.apple.com/search?term=jack+johnson')
    logger.info(f"Connected to {itunes_response.url}")
    if itunes_response.status_code != 200:
        logger.error(
            "Oops. Something went wrong trying to connect to the itunes server."
        )
        logger.error(
            f"Code: {itunes_response.status_code}, Body: {itunes_response.content}"
        )
        return None

    itunes_json_dict = json.loads(itunes_response.content)
    for index, search_result in enumerate(itunes_json_dict["results"]):
        try:
            if mode == modes.Modes.SONG:
                result = itunes_api.ItunesApiSongModel.model_validate(search_result)
                year = search_result[result.releaseDateKey].split("-")[
                    0
                ]  # will grab the year from date formatted 2016-06-01
                result.releaseDate = year
            elif mode == modes.Modes.ALBUM:
                result = itunes_api.ItunesApiAlbumKeys.model_validate(search_result)

            parsed_results_list.append(result)
        except ValidationError as e:
            logger.warning(
                f"Skipping the display of song at index [{index}] as it could not be loaded into expected format.\n{e}"
            )

    return parsed_results_list


def artwork_searcher(url: str) -> Optional[Response]:
    """Album artwork searcher.

    Args:
        url (str): the url for the artwork

    Returns:
        Response: the response for the artwork request
    """
    artwork_size_list = [
        "100x100",
        "500x500",
        "1000x1000",
        "1500x1500",
        "2000x2000",
        "2500x2500",
        "3000x3000",
    ]
    i = len(artwork_size_list) - 1

    response = requests.get(url.replace("100x100", artwork_size_list[i]))
    while response.status_code != 200 and i != 0:
        logger.info(f"- Size not found -- Trying size: {artwork_size_list[i]}")
        response = requests.get(url.replace("100x100", artwork_size_list[i]))
        i -= 1

    if i == 0:
        logger.info("Couldnt find album art. Your file wont have the art.")
        return None

    else:
        logger.info(f"Found art at size: {artwork_size_list[i]}")

    return response



from enum import Enum


class Modes(Enum):
    """enum class containing songbirdcore modes"""

    ALBUM = "album"
    """Specifies album mode"""
    SONG = "song"
    """Specifies song mode"""


def get_mode_values():
    return [e.value for e in Modes]



import pytest
from songbirdcore import youtube
import uuid
import os


@pytest.mark.parametrize(
    "video_url",
    [
        "https://www.youtube.com/watch?v=Ixrje2rXLMA",
        "https://www.youtube.com/watch?v=wOwblaKmyVw",
    ],
)
@pytest.mark.parametrize("fmt", ["m4a", "mp3"])
def test_run_download(video_url: str, fmt: str):
    res = youtube.run_download(
        url=video_url,
        file_path_no_format=os.path.join(os.sep, "tmp", str(uuid.uuid4())),
        file_format=fmt,
    )
    assert res is not None


def test_get_video_links():
    youtube_home_url: str = "https://www.youtube.com"
    youtube_search_url: str = "https://www.youtube.com/results"
    youtube_search_payload: dict = {"search_query": "jolene dolly parton"}

    link_list, links = youtube.get_video_links(
        youtube_home_url=youtube_home_url,
        youtube_search_url=youtube_search_url,
        youtube_query_payload=youtube_search_payload,
        render_timeout=20,
        render_wait=0.2,
        retry_count=3,
        render_sleep=1,
    )

    assert len(link_list) > 0 and len(links) > 0



import pytest

from typing import List, Union
import shutil
import os, sys

from songbirdcore import itunes
from songbirdcore.models import modes, itunes_api
from test_common import create_test_folder

RESOURCES_FOLDER = "resources"


@pytest.fixture()
def query_api() -> (
    List[Union[itunes_api.ItunesApiSongModel, itunes_api.ItunesApiAlbumKeys]]
):
    """pytest fixture that queries the itunes search
    api for songs.

    Yields:
        List[Union[itunes_api.ItunesApiSongModel, itunes_api.ItunesApiAlbumKeys]]: a list of songs that match the album
    """
    results = itunes.query_api(
        search_variable="jolene", limit=20, mode=modes.Modes.SONG, lookup=False
    )
    yield results


@pytest.fixture()
def query_api_album() -> (
    List[Union[itunes_api.ItunesApiSongModel, itunes_api.ItunesApiAlbumKeys]]
):
    """pytest fixture that queries the itunes
    search api for an album, and returns
    the songs in that album

    Yields:
        List[Union[itunes_api.ItunesApiSongModel, itunes_api.ItunesApiAlbumKeys]]: the list of song properties
    """
    album_results = itunes.query_api(
        search_variable="jolene", limit=20, mode=modes.Modes.ALBUM, lookup=False
    )
    # select first result by default
    result = album_results[0]
    song_results = itunes.query_api(
        search_variable=result.collectionId,
        limit=result.trackCount,
        mode=modes.Modes.SONG,
        lookup=True,
    )

    yield song_results


def test_query_api(query_api):
    assert len(query_api) > 0


def test_query_api_album(query_api_album):
    assert len(query_api_album) > 0


def test_m4a_tagger(create_test_folder, query_api):
    """test m4a tagging function executes
    successfully
    """
    input_fpath = os.path.join(sys.path[0], RESOURCES_FOLDER, "empty.m4a")
    output_fpath = os.path.join(create_test_folder, "empty.m4a")
    shutil.copy(input_fpath, output_fpath)
    # query api, get the first result
    tags = query_api[0]
    result = itunes.m4a_tagger(file_path=output_fpath, song_tag_data=tags)

    assert result == True


def test_mp3_tagger(create_test_folder, query_api):
    """test mp3 tagging function executes successfully"""
    input_fpath = os.path.join(sys.path[0], RESOURCES_FOLDER, "empty.mp3")
    output_fpath = os.path.join(create_test_folder, "empty.mp3")
    shutil.copy(input_fpath, output_fpath)
    # query api, get the first result
    tags = query_api[0]
    result = itunes.mp3ID3Tagger(mp3_path=output_fpath, song_tag_data=tags)

    assert result == True


@pytest.fixture()
def get_itunes_lib_path():
    return os.path.join(sys.path[0], RESOURCES_FOLDER, "mock-itunes-lib")


def test_itunes_lib_search_default(get_itunes_lib_path):
    """test itunes lib search functionality using mock itunes library"""
    results = itunes.itunes_lib_search(
        itunes_lib_path=get_itunes_lib_path,
        search_parameters="empty",
        album_properties=None,
    )
    assert len(results) > 0


def test_itunes_lib_search_with_props(query_api_album, get_itunes_lib_path):
    """test itunes lib search enhanced with album properties"""
    props = query_api_album[0]
    results = itunes.itunes_lib_search(
        itunes_lib_path=get_itunes_lib_path,
        search_parameters="jolene",
        album_properties=props,
    )
    assert len(results) > 0


def test_artwork_searcher(query_api):
    """test getting artwork url given song properties"""
    response = itunes.artwork_searcher(url=query_api[0].artworkUrl100)
    assert response.status_code == 200



import pytest
from songbirdcore import web


@pytest.fixture()
def get_youtube_session():
    return web.SimpleSession("youtube", root_url="https://www.youtube.com")


def test_enter_search_form(get_youtube_session: str):
    """test entering youtube's main search from"""
    session = get_youtube_session
    response = session.enter_search_form(
        search_url="https://www.youtube.com/results",
        payload={"search_query": "billy joel"},
    )
    assert response is not None and response.status_code == 200



from typing import Optional, List, Dict
import logging
from requests_html import HTMLSession
import requests
from requests import Response

logger = logging.getLogger(__name__)


class SimpleSession:
    def __init__(
        self,
        name: str,
        root_url: str,
        credentials: Optional[dict] = {},
        headers: Optional[dict] = {},
    ):
        self.name = name
        self.root_url = root_url
        self.credentials = {}
        self.headers = headers
        if headers is None:
            self.headers = {  # default header is copied from linux pc after submitting request
                "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8",
                "Accept-Encoding": "gzip, deflate, br",
                "Accept-Language": "en-US,en;q=0.5",
                "Connection": "keep-alive",
                "Host": "www.youtube.com",
                "Sec-Fetch-Dest": "document",
                "Sec-Fetch-Mode": "navigate",
                "Sec-Fetch-Site": "cross-site",
                "Upgrade-Insecure-Requests": "1",
                "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/123.0.0.0 Safari/537.36",
            }
        self.current_url = ""

        # initialize a session for this class.. to be used for all requests
        self.s = HTMLSession()

    def get_form_inputs(
        self, form_url: str, payload={}, log_calls: Optional[bool] = True
    ) -> Optional[dict]:
        """Get the inputs for a form on a webpage

        Args:
            form_url (str): the url for the form
            payload (dict, optional): the payload to insert into the form. Defaults to {}.
            log_calls (Optional[bool], optional): whether to log information within this method

        Returns:
            Optional[dict]: a dictionary containing the form
                inputs for the form_url, updated with payload
        """
        # initialize form_inputs to be empty each request
        form_inputs = {}
        try:
            form_response = self.s.get(form_url, headers=self.headers)
        except requests.exceptions.RequestException as e:
            logger.error(f"Error submitting request to: {form_url}", e)
            return None

        if log_calls:
            logger.info(f"Loaded web page: {form_response.url}!")
        # incase you want to parse through the login page.. see below comment
        inputs = form_response.html.find("input")

        # adds the csrf middleware tokens to login details.. usually stored
        # in <name> and <value> html tags
        for inputfield in inputs:
            key = inputfield.attrs["name"]
            form_inputs.update({key: ""})

        # remove None type attributes
        try:
            form_inputs.pop(None)
        except KeyError:
            logger.debug("No nonetype attributes to be removed.")

        form_inputs.update(payload)
        if log_calls:
            logger.info(f"Auto filled the web form with inputs: {form_inputs}")

        form_response.close()
        return form_inputs

    def enter_search_form(
        self,
        search_url: str,
        form_url: Optional[str] = None,
        payload: Optional[dict] = None,
        render_timeout: Optional[int] = 10,
        render_wait: Optional[float] = 0.2,
        render_sleep: Optional[int] = 1,
        log_calls: Optional[bool] = True,
    ) -> Optional[Response]:
        """
        Enter into a search form for a website
        Args:
            form_url (Optional[str]): the url for the html form. If not passed, uses the root_url for the class.
            search_url (str): the url to perform the search against
            payload (Optional[dict]): optionally include a payload for the request

        Returns:
            Optional[Response]: the rendered html response, or None if failure occurs
        """
        if form_url is None:
            form_url = self.root_url

        form_inputs = self.get_form_inputs(form_url, payload, log_calls)

        if form_inputs is None:
            return None
        try:
            response = self.s.get(search_url, params=form_inputs, headers=self.headers)
        except requests.exceptions.RequestException as e:
            logger.error(f"Error submitting request to: {search_url}", e)
            return None

        logger.debug("Rendering html for : " + response.url)
        response.html.render(
            timeout=render_timeout, wait=render_wait, sleep=render_sleep
        )
        logger.debug("Rendering complete for : " + response.url)

        response.close()
        return response

    def close(self):
        self.s.close()



import pytest
import os, sys, shutil
from pydantic import BaseModel

from songbirdcore import common


@pytest.fixture
def create_test_folder():
    tests_data_folder = os.path.join(sys.path[0], "tests-output")
    if not os.path.exists(tests_data_folder):
        os.mkdir(tests_data_folder)
    yield tests_data_folder
    # test cleanup
    shutil.rmtree(tests_data_folder)


def test_load_data():
    toml_path = os.path.join(sys.path[0], "..", "..", "pyproject.toml")
    data = common.load_toml(toml_path)
    assert data is not None


def test_set_logger_config_globally():
    try:
        common.set_logger_config_globally()
    except Exception as e:
        pytest.fail(f"Unexpected exception {e}!")


def test_name_plate():
    try:
        common.name_plate(entries=["foo"])
    except Exception as e:
        pytest.fail(f"Unexpected exception {e}!")


@pytest.fixture()
def create_dummy_file(create_test_folder):
    fname = os.path.join(create_test_folder, "dummy.txt")

    with open(fname, "w") as f:
        f.write("hello")
    yield fname


def fname_duper(create_dummy_file):

    fname = common.fname_duper(fname=create_dummy_file, limit=2, count=1, dup_key="dup")

    assert fname is not None


def test_fname_duper_fails(create_dummy_file):
    fname = common.fname_duper(fname=create_dummy_file, limit=1, count=1, dup_key="dup")

    assert fname is None


def test_remove_illegal_characters():
    """test remove illegal character"""
    try:
        result = common.remove_illegal_characters("\\\"/*?<>|':")
    except Exception as e:
        pytest.fail(f"Unexpected exception {e}!")

    assert result == ""


def test_find_file(create_dummy_file):
    """test function for finding a file"""
    result = common.find_file(
        path=os.path.dirname(create_dummy_file),
        filename=os.path.basename(create_dummy_file),
    )

    assert len(result) > 0


def test_pretty_list_of_basemodel_printer(capsys):
    """test function runs without exception"""

    class Model(BaseModel):
        foo: str = "bar"

    try:
        common.pretty_list_of_basemodel_printer(list_of_models=[Model()])
    except Exception as e:
        pytest.fail(f"Unexpected exception {e}!")


def test_pretty_list_of_baseModel_printer(capsys):
    """test ignore_keys functionality"""

    class Model(BaseModel):
        foo: str = "bar"
        ignore_me: str = "ignore_me"

    try:
        common.pretty_list_of_basemodel_printer(
            list_of_models=[Model()], ignore_keys=["ignore_me"]
        )
    except Exception as e:
        pytest.fail(f"Unexpected exception {e}!")

    captured = capsys.readouterr()
    assert "ignore_me" not in captured.out


def tests_pretty_lst_printer():
    """test function runs without exception"""
    try:
        common.pretty_lst_printer(["hi", "there"])
    except Exception as e:
        pytest.fail(f"Unexpected exception: {e}")



import os, sys

from googleapiclient.discovery import build
from google_auth_oauthlib.flow import InstalledAppFlow
from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from googleapiclient import http

import logging
from typing import Optional

logger = logging.getLogger(__name__)


def save_song(
    gdrive_folder_id: str,
    credentials_path: str,
    token_path: str,
    song_name: str,
    song_path: str,
    auth_port: int,
    bind_addr: Optional[str] = None,
) -> str:
    """save a song to google drive

    Args:
        gdrive_folder_id (str): google drive folder id
        credentials_path (str): path to a credentials.json file
        token_path (str): path to a google cloud token file
        song_name (str): the name of the song (filename)
        song_path (str): the path of the song locally
        auth_port (int): the port to use for oauth
        bind_addr (Optional[str], optional): optionally specify the bind address, otherwise localhost is used. Defaults to None.

    Returns:
        str: the file id of the upload
    """
    creds = None
    # The file token.json stores the user's access and refresh tokens, and is
    # created automatically when the authorization flow completes for the first
    # time.
    scopes = ["https://www.googleapis.com/auth/drive"]
    if os.path.exists(token_path):
        creds = Credentials.from_authorized_user_file(token_path, scopes)
    # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            flow = InstalledAppFlow.from_client_secrets_file(credentials_path, scopes)
            creds = flow.run_local_server(port=auth_port, bind_addr=bind_addr)
        # Save the credentials for the next run
        with open(token_path, "w") as token:
            token.write(creds.to_json())

    logger.debug(f"Loaded credentials: {creds}")
    service = build("drive", "v3", credentials=creds)

    file_metadata = {"name": song_name, "parents": [gdrive_folder_id]}
    media = http.MediaFileUpload(song_path, mimetype="audio/jpeg")
    file = (
        service.files()
        .create(body=file_metadata, media_body=media, fields="id")
        .execute()
    )
    file_id = file.get("id")
    logger.info(f"File creation successful -- ID: {file_id}")
    return file_id


if __name__ == "__main__":
    fname = "test.txt"
    fpath = os.path.join(sys.path[0], fname)
    cred_path = os.path.join(sys.path[0], "data", "gdrive", "credentials.json")
    token_path = os.path.join(sys.path[0], "data", "gdrive", "token.json")
    fid = input("Enter folder id: ")
    with open(fpath, "w") as f:
        f.write("swag")
    save_song(fid, cred_path, token_path, "test.txt", fpath, 8080)
    os.remove(fpath)



from typing import List
import logging
import glob
import os, sys
import tomllib

from typing import Optional
import logging
from pydantic import BaseModel

from .models import itunes_api, modes
from .version import version

logger = logging.getLogger(__name__)


def load_toml(toml_path: str):
    with open(toml_path, "rb") as f:
        data = tomllib.load(f)
    return data


def set_logger_config_globally(log_level=logging.INFO) -> None:
    """Sets the python logging module settings for output
    to stdout and to file.

    Args:
        log_level (str): the log level

    """
    logging.basicConfig(
        level=log_level,
        format="[%(levelname)s] %(name)s: %(message)s",
        handlers=[logging.StreamHandler()],
    )


def name_plate(entries: List[str]) -> None:
    """renders the songbird entrypoint name plate

    Args:
        entries (List[str]): add additional entries to the nameplate via this list
    """
    # load project version from file
    print("===============================")
    print("=----Welcome to songbird🐦----=")
    print("===============================")
    print(f"--songbirdcore {version}")
    for entry in entries:
        print(entry)
    print("===============================")
    print("Message from developer:")
    print(
        "\t- dependencies have been upgraded, make sure you have run 'playwright install'."
    )
    print(
        "\t- If you encounter errors, please create an issue here https://github.com/cboin1996/songbirdcore/issues/"
    )
    print(
        f"At the main menu, type one of {[mode.value for mode in modes.Modes]} to switch modes!"
    )


def fname_duper(fname: str, limit: int, count: int, dup_key: str) -> Optional[str]:
    """Generates a duplicate filename for when a filename already exists

    Args:
        fname (str): filename
        limit (int): a limit of dups before quitting the attempt
        count (int): recursive count tracker
        dup_key (str): the key to use as the duplicate addon

    Returns:
        Optional[str]: the modified filename, or None if the limit has been reached.
    """
    fname_split = os.path.splitext(fname)
    fname_noext = fname_split[0]
    ext = fname_split[1]
    if count == limit:
        logger.error(
            f"Max retry limit {limit} reached for fname {fname}. Please try changing some filenames and try again later."
        )
        return None
    if os.path.exists(fname):
        fname = fname_duper(fname_noext + dup_key + ext, limit, count + 1, dup_key)

    return fname


def remove_illegal_characters(filename) -> str:
    """
    Used for stripping file names of illegal characters used for saving

    Args:
        filename (str): the file's name to strip illegal characters from

    Returns:
        str: stripped file name
    """
    return (
        filename.replace("\\", "")
        .replace('"', "")
        .replace("/", "")
        .replace("*", "")
        .replace("?", "")
        .replace("<", "")
        .replace(">", "")
        .replace("|", "")
        .replace("'", "")
        .replace(":", "")
    )


def find_file(path: str, filename: str) -> List[str]:
    """Simple glob search for a file

    Args:
        path (str): the path to the root folder to search within
        filename (str): the filename (supports glob patterns)

    Returns:
        List[str]: list of paths found that match
    """
    paths = glob.glob(os.path.join(path, filename))
    return paths


def pretty_list_of_basemodel_printer(
    list_of_models: List[BaseModel], ignore_keys: Optional[List[str]] = None
):
    """
    renders a list to stdio given a list of pydantic BaseModel objects

    Args:
        list_of_models (List[BaseModel]): list of pydantic models to print
        ignore_keys (Optional[List[str]], optional): any keys/fields in BaseModel not to print
    """
    i = len(list_of_models) - 1
    logger.info("------------------------")
    for element in reversed(list_of_models):
        logger.info(i)
        for k, v in element.model_dump().items():
            if ignore_keys is not None:
                if (
                    k not in ignore_keys
                ):  # print the key and value if not in ignore_keys or special_dict
                    print("\t%s - %s" % (k, v))

            else:  # (default case) print the key and value
                print("\t%s - %s" % (k, v))
        i -= 1
        print("------------------------")


def pretty_lst_printer(lyst: List):
    """print a list to stdio

    Args:
        lyst (List): the list to print
    """
    for idx, item in enumerate(lyst):
        logger.info(f"\t [{idx}] - {item}")



import logging
from typing import Optional
from typing import Tuple, List
import requests_html
from bs4 import BeautifulSoup
import os, sys

from .models import itunes_api
from . import web
from . import common

logger = logging.getLogger(__name__)
import yt_dlp as youtube_dl
from yt_dlp.utils import DownloadError, UnavailableVideoError, ExtractorError


def get_video_links(
    youtube_home_url: str,
    youtube_search_url: str,
    youtube_query_payload: dict,
    render_timeout: int,
    render_wait: float,
    retry_count: Optional[int] = 3,
    render_sleep: Optional[int] = 1,
) -> Tuple[List[str], List[requests_html.Element]]:
    """
    Args:
        render_timeout (int): amount of time before abandoning a render
        render_wait (float): the amount of time before attempting a render
        retry_count (int): the number of retries for a render
        render_sleep (Optional[int]): the amount of time to wait after rendering
    """
    headers = {
        "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/106.0.0.0 Safari/537.36",  # ubuntu chrome headers
        "Referrer": youtube_home_url,
    }
    session = web.SimpleSession("youtube", root_url=youtube_home_url, headers=headers)
    tries = 0
    # Only need to log info about these requests on first try to simplify UI experience
    log_attempts = True
    while tries < retry_count:
        # First, enter the search form on the youtube home page
        response = session.enter_search_form(
            search_url=youtube_search_url,
            payload=youtube_query_payload,
            render_timeout=render_timeout,
            render_wait=render_wait,
            render_sleep=render_sleep,
            log_calls=log_attempts,
        )
        if response == None:
            logger.error(
                f"Error occurred performing a search for {youtube_query_payload} against url {youtube_search_url}. Please try again."
            )
            tries += 1
            log_attempts = False
            continue
        # Get the list of hrefs to each video on the home page
        links = response.html.find("#video-title")
        if len(links) == 0:
            logger.warning(f"{tries+1}:{retry_count}.")
            tries += 1
            log_attempts = False
        else:
            break
    session.close()
    if tries >= retry_count:
        logger.error(
            f"Failed to get links from {youtube_home_url} after {retry_count} tries."
        )
        return None, None

    link_list = []
    # create a user friendly list, containing videos with title and href refs.
    for idx, link in enumerate(links):
        if "title" in link.attrs and "href" in link.attrs:
            link_list.append(
                f"{link.attrs['title']} - {youtube_home_url+link.attrs['href']}"
            )
        # the actual raw list should match list size so we can properly select the link after.
        else:
            links.pop(idx)

    return link_list, links


class YtDlLogger(object):
    """
    Used for setting up youtube_dl logging
    """

    def info(self, msg):
        logger.info(msg)

    def debug(self, msg):
        logger.debug(msg)

    def warning(self, msg):
        logger.warning(msg)

    def error(self, msg):
        logger.error(msg)


def my_hook(d):
    """
    Hook for youtube_dl
    Args:
        d: download object from youtube_dl
    """
    if d["status"] == "finished":
        sys.stdout.write("\n")
        logger.info("Done downloading, now converting ...")

    if d["status"] == "downloading":
        p = d["_percent_str"]
        p = p.replace("%", "")
        song_name = d["filename"].split(os.sep)[-1]
        sys.stdout.write(
            f"\rDownloading to file: {song_name}, {d['_percent_str']}, {d['_eta_str']}"
        )
        sys.stdout.flush()
    if d["status"] == "error":
        logger.error("Error occured during download.")
        success_downloading = False


def run_download(url: str, file_path_no_format: str, file_format: str) -> str:
    """Run a download from youtube

    Args:
        url (str): the url to download
        file_path_no_format (str): the file path excluding file format
        file_format (str): the file format
    Returns:
        str: the filepath.
    """
    local_file_path = f"{file_path_no_format}.{file_format}"
    ydl_opts = {
        "format": "bestaudio/best",
        "cachedir": False,
        "postprocessors": [
            {
                "key": "FFmpegExtractAudio",
                "preferredcodec": file_format,
                "preferredquality": "192",
            }
        ],
        "nocheckcertificate": True,
        "logger": YtDlLogger(),
        "progress_hooks": [my_hook],
        "outtmpl": file_path_no_format + ".%(ext)s",
    }
    try:
        with youtube_dl.YoutubeDL(ydl_opts) as ydl:
            ydl.download(url)
        logger.info(f"Downloading successful. File stored locally: {local_file_path}")
        return local_file_path
    except Exception as e:
        logger.exception(f"Failed to complete the download of song at url: {url}.")
        return None



from pydantic import BaseModel
from typing import Optional, Union


class ItunesApiSongModel(BaseModel):
    trackName: str
    """specifies song name"""
    artistName: str
    """specifies song artist name"""
    collectionName: str
    """specifies album name"""
    artworkUrl100: str
    """specifies url pointing to album artwork"""
    primaryGenreName: str
    """specifies genre name for the song"""
    trackNumber: int
    """specifies song number in album"""
    trackCount: int
    """specifies number of songs in album"""
    collectionId: Union[int, str] = ""
    """specifies id for the album"""
    collectionArtistName: str = ""
    """specifies artist name for the album"""
    discNumber: int
    """specifies disc number in album"""
    discCount: int
    """specifies disc count in album"""
    releaseDate: str
    """specifies release date of album"""
    releaseDateKey: str = "releaseDate"
    """specifies key expected for releaseDate field in itunes api response object"""


class ItunesApiAlbumKeys(BaseModel):
    artistName: str
    """ specifies artist name for album"""
    collectionName: str
    """ specifies artist name for collection"""
    trackCount: int
    """ specifies number of songs in album"""
    collectionId: Union[int, str] = ""
    """ specified itunes api collection id"""
